/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2010 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ClassLoaderService.java 22296 2012-03-23 14:08:06Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.admin.classloader.service;

import java.io.IOException;
import java.io.Serializable;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import javax.management.ObjectName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;




/**
 * Service that is allowing to deal with JOnAS classloaders <br />
 * This includes System/OSGi classloaders and Java EE module classloaders
 * @author Florent Benoit
 */
public class ClassLoaderService extends JMXHelperService implements Serializable {

    /**
     * Logger of the class
     */
    private Log logger = LogFactory.getLog(ClassLoaderService.class);

    /**
     * Serial Version UID.
     */
    private static final long serialVersionUID = 5509445326180411783L;

    /**
     * J2EE server ObjectName.
     */
    private ObjectName javaEEServerObjectName = null;


    /**
     * Default constructor.
     */
    public ClassLoaderService() {
        super();

        // J2EE server MBean
        this.javaEEServerObjectName = buildObjectName(getDomainName() + ":j2eeType=J2EEServer,name=" + getServerName());
    }

    public void applyAttributes(final Document document, final String xPathExpression, final String attributeName, final String attributeValue) {
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xPath = xPathFactory.newXPath();

        XPathExpression xPathExpr = null;
        try {
            xPathExpr = xPath.compile(xPathExpression);
        } catch (XPathExpressionException e) {
            throw new IllegalStateException("Cannot analyze configuration file", e);
        }
        NodeList nodeList = null;
        try {
            nodeList = (NodeList) xPathExpr.evaluate(document, XPathConstants.NODESET);
        } catch (XPathExpressionException e) {
            throw new IllegalStateException("Cannot analyze document", e);
        }
        if (nodeList != null) {
            for (int i = 0; i < nodeList.getLength(); i++) {
                Element element = (Element) nodeList.item(i);
                element.setAttribute(attributeName, attributeValue);
            }
        }
    }


    /**
     * Gets the exported packages and the bundles for a given resource name (class or package) depending on the isClass flag.
     * @param name the resource name
     * @param isClass true if the resource is a class, else a package
     * @return the exported packages in a XML document.
     */
    public Document getSystemExportedPackages(final String name, final boolean isClass) {
        String stringExportedPackages = (String) invoke(javaEEServerObjectName, "getExportedPackages", new Object[] {name, isClass}, new String[] {String.class.getName(), boolean.class.getName()});
        Document document = transformXMLIntoDocument(stringExportedPackages);

        // Needs to adds some Flex datagrid attributes on the resulting XML document


        // First, set the isBranch attribute to false on some nodes
        applyAttributes(document, "//exported-packages/exported-package/exporting-bundle | //exported-packages/exported-package/exporting-bundle/* | //exported-packages/exported-package/importing-bundles/*", "isBranch", "false" );

        // Then set attribute name for bundles
        applyAttributes(document, "//exported-packages/exported-package/importing-bundles", "name", "importing-bundles" );
        applyAttributes(document, "//exported-packages/exported-package/importing-bundles/importing-bundle", "name", "importing-bundle" );
        applyAttributes(document, "//exported-packages/exported-package/exporting-bundle", "name", "exporting-bundle" );

        return document;
    }


    public Document objectNameloadClass(final String objectNameString, final String className) {
        // Check if objectName exists
        ObjectName objectName = buildObjectName(objectNameString);

        if (!isRegistered(objectName)) {
            // Nothing exists, return empty document
            throw new IllegalStateException("ObjectName '" + objectNameString + "' doesn't exists");
        }

        String stringLoadClass = (String) invoke(objectName, "loadClass", new Object[] {className}, new String[] {String.class.getName()});
        return transformXMLIntoDocument(stringLoadClass);
    }


    /**
     * Gets the path of a given resource name by asking system classloader
     * @param resourceName the name of the resource
     * @return the URLs list
     */
    public String[] objectNameGetResources(final String objectNameString, final String resourceName) {
        // Check if objectName exists
        ObjectName objectName = buildObjectName(objectNameString);

        if (!isRegistered(objectName)) {
            // Nothing exists, return empty document
            throw new IllegalStateException("ObjectName '" + objectNameString + "' doesn't exists");
        }

        URL[] urls = (URL[]) invoke(objectName, "getResources", new Object[] {resourceName}, new String[] {String.class.getName()});

        String[] urlsString = new String[urls.length];
        int i = 0;
        for (URL url : urls) {
            urlsString[i++] = url.toExternalForm();
        }

        return urlsString;
    }

    /**
     * Gets the path of a given resource name by asking system classloader
     * @param resourceName the name of the resource
     * @return the URLs list
     */
    public String[] getSystemResources(final String resourceName) {
        URL[] urls = (URL[]) invoke(javaEEServerObjectName, "getResources", new Object[] {resourceName}, new String[] {String.class.getName()});

        String[] urlsString = new String[urls.length];
        int i = 0;
        for (URL url : urls) {
            urlsString[i++] = url.toExternalForm();
        }

        return urlsString;
    }


    /**
     * @param objectName the ObjectName of the classloader MBean
     * @return the filters that are applied on the given classloader/ ObjectName
     */
    public String[] getClassLoaderFilters(final ObjectName objectName) {
        String[] filters = (String[]) getAttribute(objectName, "classLoaderFilters");
        return filters;
    }

    /**
     * Add XML filters for the given objectname by using the given document and parent element.
     * @param document the XML document to use
     * @param parentElement the XML element which is the parent element
     * @param objectName the objectname of the MBean
     */
    public void addFilters(final Document document, final Element parentElement, final ObjectName objectName) {
        String[] filters = null;
        try {
            filters = getClassLoaderFilters(objectName);
        } catch (IllegalStateException e) {
            // no filters, return
            return;
        }

        Element filtersElement = document.createElement("filters");
        parentElement.appendChild(filtersElement);

        // Add each filter name
        if (filters != null) {
            for (String filter : filters) {
                Element filterElement = document.createElement("filter");
                filtersElement.appendChild(filterElement);
                Text text = document.createTextNode(filter);
                filterElement.appendChild(text);
            }
        }
    }


    /**
     * Transform XML string into a XML document.
     * @param xmlEntry the given string
     * @return a XML document
     */
    protected Document transformXMLIntoDocument(final String xmlEntry) {
        // Create builder with factory
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new IllegalStateException("Cannot build document builder", e);
        }

        // Create Document
        Document document = null;

        // Parse XML
        try {
            document = builder.parse(new InputSource(new StringReader(xmlEntry)));
        } catch (SAXException e) {
            throw new IllegalStateException("Cannot handle stream", e);
        } catch (IOException e) {
            throw new IllegalStateException("Cannot handle stream", e);
        }



        // Log ?
        if (logger.isDebugEnabled()) {
            StringWriter stringWriter = new StringWriter();
            StreamResult streamResult = new StreamResult(stringWriter);

            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer;
            try {
                transformer = transformerFactory.newTransformer();
            } catch (TransformerConfigurationException e) {
                throw new IllegalStateException("Cannot get a transformer", e);
            }

            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");

            // transform OUTPUT
            try {
                transformer.transform(new DOMSource(document), streamResult);
            } catch (TransformerException e) {
                throw new IllegalStateException("Cannot transform the document", e);
            }

            logger.debug("writing results of XML : ", stringWriter.toString());
        }

        // display result
        return document;
    }

    /**
     * Gets latest part of name.
     * @param name the value to shorter
     * @return the modified value
     */
    protected static String stripName(final String name) {
        String newName = name;

        // Directory
        if (name.endsWith("/")) {
            newName = name.substring(0, name.length() - 1);
        }


        int i = newName.lastIndexOf('/');
        if (i != -1) {
            newName = newName.substring(i + 1);
        }
        return newName;
    }


    protected List<ApplicationsInfo> getModules(final ObjectName objectName, final String filter, final String suffix) {
        List<ApplicationsInfo> list = new ArrayList<ApplicationsInfo>();

        Set<ObjectName> objectMBeans = null;
        try {
            objectMBeans = getMBeanServerConnection().queryNames(objectName, null);
        } catch (IOException e) {
            throw new IllegalStateException("Cannot get ObjectName", e);
        }

        for (ObjectName oName : objectMBeans) {
            String name = oName.getKeyProperty("name");
            // strip the first / if any
            name = stripName(name);

            if (filter != null && !"".equals(filter)) {
                if (!name.toLowerCase().contains(filter.toLowerCase())) {
                    continue;
                }
            }

            ApplicationsInfo appInfo = new ApplicationsInfo();
            appInfo.setName(name + suffix);
            appInfo.setObjectName(oName.toString());
            list.add(appInfo);
        }

        return list;
    }



    /**
     * Return a list of applications based on a given filtering search.
     * @param filter a text filtering
     * @param apps if EAR are included
     * @param ejbs if EJBJAR are included
     * @param webApps if WAR are included
     * @return a list of applications
     */
    public List<ApplicationsInfo> getApplications(final String filter, final boolean apps, final boolean ejbs, final boolean webApps) {

        List<ApplicationsInfo> list = new ArrayList<ApplicationsInfo>();

        // Get EAR applications
        if (apps) {
            list.addAll(getModules(buildObjectName(getDomainName() + ":j2eeType=J2EEApplication,J2EEServer=" + getServerName() + ",name=*"), filter, "[EAR]"));
        }

        if (ejbs) {
            list.addAll(getModules(buildObjectName(getDomainName() + ":j2eeType=EJBModule,J2EEServer=" + getServerName() + ",J2EEApplication=null,name=*"), filter, "[EJB]"));
        }

        if (webApps) {
            list.addAll(getModules(buildObjectName(getDomainName() + ":j2eeType=WebModule,J2EEServer=" + getServerName() + ",J2EEApplication=null,name=*"), filter, "[WEB]"));
        }

        // Sort by Name
        Collections.sort(list);

        return list;

    }


    protected void addEJBbClassLoaderElementsInfo(final Document document, final ObjectName objectName, int id,
        final int parentId, final Element root, final boolean sharing) {
        int elementID = id++;
        // Add web module information

        // create a link between parent node and ourself
        Element edgeElement = document.createElement("Edge");
        edgeElement.setAttribute("fromID", String.valueOf(parentId));
        edgeElement.setAttribute("toID", String.valueOf(elementID));
        edgeElement.setAttribute("edgeLabel", "");
        edgeElement.setAttribute("flow", "1.0");
        edgeElement.setAttribute("color", "0x000000");
        root.appendChild(edgeElement);

        Element ejbModuleElement = document.createElement("Node");
        root.appendChild(ejbModuleElement);

        ejbModuleElement.setAttribute("id", String.valueOf(elementID));

        // Write objectName
        Element objectNameElement = document.createElement("object-name");
        ejbModuleElement.appendChild(objectNameElement);
        Text text = document.createTextNode(objectName.toString());
        objectNameElement.appendChild(text);


        if (sharing) {
            ejbModuleElement.setAttribute("name", "All EJBs");
            ejbModuleElement.setAttribute("toolTips", "All EJBs");
        } else {
            // write name
            URL url = (URL) getAttribute(objectName, "url");
            ejbModuleElement.setAttribute("name", stripName(url.toString()));
            ejbModuleElement.setAttribute("toolTips", stripName(url.toString()));

            // Add filters
            addFilters(document, ejbModuleElement, objectName);

        }

        // Write Java delegation Model
        ejbModuleElement.setAttribute("nodeIcon", "3");

    }

    protected void addWebClassLoaderElementsInfo(final Document document, final ObjectName objectName, int id,
        final int parentId, final Element root) {
        int elementID = id++;
        // Add web module information

        // create a link between parent node and ourself
        Element edgeElement = document.createElement("Edge");
        edgeElement.setAttribute("fromID", String.valueOf(parentId));
        edgeElement.setAttribute("toID", String.valueOf(elementID));
        edgeElement.setAttribute("edgeLabel", "");
        edgeElement.setAttribute("flow", "1.0");
        edgeElement.setAttribute("color", "0x000000");
        root.appendChild(edgeElement);

        Element webModuleElement = document.createElement("Node");
        root.appendChild(webModuleElement);

        webModuleElement.setAttribute("id", String.valueOf(elementID));

        // Write objectName
        Element objectNameElement = document.createElement("object-name");
        webModuleElement.appendChild(objectNameElement);
        Text text = document.createTextNode(objectName.toString());
        objectNameElement.appendChild(text);

        // write name
        URL warURL = (URL) getAttribute(objectName, "warURL");
        webModuleElement.setAttribute("name", stripName(warURL.toString()));
        webModuleElement.setAttribute("toolTips", stripName(warURL.toString()));

        // Write Java delegation Model
        boolean java2DelegationModel = ((Boolean) getAttribute(objectName, "java2DelegationModel")).booleanValue();
        webModuleElement.setAttribute("delegation-model", Boolean.toString(java2DelegationModel));
        if (java2DelegationModel) {
            webModuleElement.setAttribute("nodeIcon", "5");
        } else {
            webModuleElement.setAttribute("nodeIcon", "6");
        }

        // Add filters
        addFilters(document, webModuleElement, objectName);


    }


    /**
     * Gets an XML document representing all the classloader data for a given Java EE application.
     * @param objectName the objectname of the EE application
     * @return a XML document
     */
    public Document getClassLoaderInfo(final String objectNameString) {

        // Create XML document...

        // Create builder with factory
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new IllegalStateException("Cannot build document builder", e);
        }

        // Create Document
        Document document = builder.newDocument();

        // Append root element we're building a Graph XML object
        Element root = document.createElement("Graph");
        document.appendChild(root);


        // Check if objectName exists
        ObjectName objectName = buildObjectName(objectNameString);

        if (!isRegistered(objectName)) {
            // Nothing exists, return empty document
            return document;
        }

        // Id counter
        int id = 0;
        int parentId = id;

        // Then needs to find data about system classloader

        Element systemClassLoaderElement = document.createElement("Node");
        systemClassLoaderElement.setAttribute("id", String.valueOf(id++));
        systemClassLoaderElement.setAttribute("name", "System\nClassLoader");
        systemClassLoaderElement.setAttribute("nodeIcon", "2");

        root.appendChild(systemClassLoaderElement);

        // Write objectName
        Element sysObjectNameElement = document.createElement("object-name");
        systemClassLoaderElement.appendChild(sysObjectNameElement);
        Text textObjectName = document.createTextNode(javaEEServerObjectName.toString());
        sysObjectNameElement.appendChild(textObjectName);

        // Add filters
        addFilters(document, systemClassLoaderElement, javaEEServerObjectName);

        // Ear type ?
        // needs to get EAR infos
        if ("J2EEApplication".equals(objectName.getKeyProperty("j2eeType"))) {

            // Name of the EAR application
            String earName = objectName.getKeyProperty("name");


            // add EAR info
            int elementID = id++;

            // Add web module information

            // create a link between parent node and ourself
            Element edgeElement = document.createElement("Edge");
            edgeElement.setAttribute("fromID", String.valueOf(parentId));
            edgeElement.setAttribute("toID", String.valueOf(elementID));
            edgeElement.setAttribute("edgeLabel", "");
            edgeElement.setAttribute("flow", "1.0");
            edgeElement.setAttribute("color", "0x000000");
            root.appendChild(edgeElement);


            Element earModuleElement = document.createElement("Node");
            root.appendChild(earModuleElement);

            earModuleElement.setAttribute("id", String.valueOf(elementID));

            // Write objectName
            Element objectNameElement = document.createElement("object-name");
            earModuleElement.appendChild(objectNameElement);
            Text text = document.createTextNode(objectName.toString());
            objectNameElement.appendChild(text);

            // write name
            URL earURL = (URL) getAttribute(objectName, "earUrl");
            earModuleElement.setAttribute("name", stripName(earURL.toString()));
            earModuleElement.setAttribute("toolTips", stripName(earURL.toString()));

            // Write Java delegation Model
            earModuleElement.setAttribute("nodeIcon", "7");

            // Add filters
            addFilters(document, earModuleElement, objectName);



            // Change parent ID
            parentId = elementID;


            // Then EJB Module (only one)
            ObjectName ejbModulesObjectName = buildObjectName(getDomainName() + ":j2eeType=EJBModule,J2EEServer=" + getServerName() + ",J2EEApplication=" + earName + ",name=*");
            Set<ObjectName> ejbModulesObjectMBeans = null;
            try {
                ejbModulesObjectMBeans = getMBeanServerConnection().queryNames(ejbModulesObjectName, null);
            } catch (IOException e) {
                throw new IllegalStateException("Cannot get ObjectName", e);
            }

            // Take the first one if any
            if (ejbModulesObjectMBeans != null  && ejbModulesObjectMBeans.size() > 0) {
                addEJBbClassLoaderElementsInfo(document, ejbModulesObjectMBeans.iterator().next(), id, parentId, root, true);
                parentId = id;
                id++;
            }


            // Then all Web modules
            ObjectName webModulesObjectName = buildObjectName(getDomainName() + ":j2eeType=WebModule,J2EEServer=" + getServerName() + ",J2EEApplication=" + earName + ",name=*");
            Set<ObjectName> webModulesObjectMBeans = null;
            try {
                webModulesObjectMBeans = getMBeanServerConnection().queryNames(webModulesObjectName, null);
            } catch (IOException e) {
                throw new IllegalStateException("Cannot get ObjectName", e);
            }

            if (webModulesObjectMBeans != null) {
                for (ObjectName webModuleObjectName : webModulesObjectMBeans) {
                    addWebClassLoaderElementsInfo(document, webModuleObjectName, id, parentId, root);
                }

            }



        }

        if ("WebModule".equals(objectName.getKeyProperty("j2eeType"))) {
            addWebClassLoaderElementsInfo(document, objectName, id, parentId, root);
        }

        if ("EJBModule".equals(objectName.getKeyProperty("j2eeType"))) {
            addEJBbClassLoaderElementsInfo(document, objectName, id, parentId, root, false);
        }


        StringWriter stringWriter = new StringWriter();
        StreamResult streamResult = new StreamResult(stringWriter);

        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        Transformer transformer;
        try {
            transformer = transformerFactory.newTransformer();
        } catch (TransformerConfigurationException e) {
            throw new IllegalStateException("Unable to get a new transformer", e);
        }

        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");

        // transform OUTPUT
        try {
            transformer.transform(new DOMSource(document), streamResult);
        } catch (TransformerException e) {
            throw new IllegalStateException("Unable to transform the document", e);
        }

        return document;
    }



}
